#include <stdio.h>
#include <stdlib.h>

#define MAX_AREA 0x43

char * checksum(int level, char * pName)
{
    char * chksum = (char *) malloc(4);
    chksum[0] = (level & 0xff);
    chksum[1] = 0;
    int i,c = level;
    for(i = 7; i >= 0; i--) c += (pName[i] ^ 0xff);
    chksum[2] = (c & 0xff);
    chksum[3] = ((c & 0xff00) >> 8);   
    return chksum;
}

char * id2Icon(char * password, int len) {
    int i;     
    char * _password = (char *)malloc(len + 1);          
    for(i = 0; i < len; i++) {
        if (password[i] == 0) _password[i] = ' ';
        if (password[i] == 1) _password[i] = 'A';
        if (password[i] == 2) _password[i] = 'W';
        if (password[i] == 3) _password[i] = 'H';
    }
    _password[len] = 0;
    return _password;
}

int main(int argc, char *argv[])
{
    printf("Super Castlevania IV password generator\nby: TFG (tfg@panicus.org)\n\n");
    
    // Player's name
    char pName[8] = {0,0,0,0,0,0,0,0};
    char asciiName[8] = {0};
    printf("Enter your name (8 letters or less): ");
    scanf("%8s", pName);
    strcpy(asciiName,pName);
    
    // De-ASCII
    int i;
    for(i = 0; i < 8; i++) {
        if(pName[i] == ' ') pName[i] = 0;
        else if((pName[i] >= 49)&&(pName[i] <= 57)) pName[i] -= 47;
        else if((pName[i] >= 65)&&(pName[i] <= 90)) pName[i] -= 54;
        else if((pName[i] >= 97)&&(pName[i] <= 122)) pName[i] -= 86;
        else pName[i] = 0;
    }
    
    // Checksum nibble table
    char cntTable[] = {0x02,0x03,0x01,0x00,0x03,0x02,0x02,0x01,
                       0x03,0x01,0x00,0x00,0x01,0x02,0x03,0x00,
                       0x01,0x01,0x02,0x00,0x03,0x03,0x00,0x02,
                       0x00,0x01,0x02,0x02,0x00,0x03,0x01,0x03};    
    
    char areaMask[] = {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
                       0x08,0x08,0x08,0x08,0x0C,0x0C,0x0C,0x0C,
                       0x0C,0x0C,0x12,0x12,0x12,0x12,0x12,0x12,
                       0x18,0x18,0x1A,0x1A,0x1A,0x1A,0x1A,0x1A,
                       0x1A,0x1A,0x1A,0x23,0x23,0x23,0x23,0x23,
                       0x23,0x23,0x2A,0x2A,0x2A,0x2A,0x2E,0x2E,
                       0x2E,0x2E,0x2E,0x2E,0x2E,0x2E,0x2E,0x37,
                       0x37,0x37,0x37,0x37,0x3C,0x3C,0x3C,0x3F,
                       0x40,0x41,0x42,0x42};
    
    char password[16] = {0,0,0,0,
                         0,0,0,0,
                         0,0,0,0,
                         0,0,0,0};
                            
    int j,k,l,m,n,o,p,q;    
    char * chksum;
    FILE * output;
    
    // Previous level indicator
    n = 0xff;
    q = 0;
    
    if((output = fopen("scv4pg_output.txt","w")) != NULL) {
        // File header
        fprintf(output,"Super Castlevania IV password generator\nby: TFG (tfg@panicus.org)\n\n");
        fprintf(output,"A = Axe, W = Holy Water, H = Heart\n");
        fprintf(output,"Name: %s",asciiName);
        
        // Generate passwords
        for(o = 0; o <= MAX_AREA; o++) {
            i = areaMask[o];
              
            // Did we already process this level?
            if(i != n) {              
                // "Human" level counter
                q++;
                
                // Calculate checksum
                chksum = checksum(i,pName);

                // "Encode"
                for(j = 0, p = 0; j < 0x20; j+=2, p++) {
                    k = cntTable[j];
                    l = chksum[k];
                    k = cntTable[j + 1];
                    l >>= (k * 2);
                    password[p] = (l & 3);
                }
                           
                // Output      
                char * pass = id2Icon(password,16);
                fprintf(output,"\n\nLevel %2d\n--------\n",q);
                for(m = 0; m < 4; m++)
                fprintf(output,"[%c][%c][%c][%c]\n",pass[m * 4],pass[(m * 4) + 1],
                                                    pass[(m * 4) + 2], pass[(m * 4) + 3]);                                                  
            }
            
            n = i;
        }
        
        // Close        
        fprintf(output,"\neof\n");
        printf("\nPassword generation logged to \'scv4pg_output.txt\'\n");
        fclose(output);
    }
    else {
        printf("\nCouldn't open \'scv4pg_output.txt\' Aborting...\n");
    }
    
    
    printf("\nPress any key to continue...");
    getch();
  
    return 0;
}
